/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c - PENTA: almost obsolete now

#include "quakedef.h"

int	r_dlightframecount;

/*
==================
R_AnimateLight
==================
*/
void R_AnimateLight (void)
{
    int			j, k;
    float		l;
    int			flight;
    int			clight;
    float		lerpfrac;
    float		backlerp;

    // light animations
    // 'm' is normal light, 'a' is no light, 'z' is double bright
    flight = (int) floor (cl.time * 10);
    clight = (int) ceil (cl.time * 10);
    lerpfrac = (cl.time * 10) - flight;
    backlerp = 1.0f - lerpfrac;

    for (j = 0; j < MAX_LIGHTSTYLES; j++)
    {
        if (!cl_lightstyle[j].length)
        {
            // was 256, changed to 264 for consistency
            d_lightstylevalue[j] = 264;
            continue;
        }
        else if (cl_lightstyle[j].length == 1)
        {
            // single length style so don't bother interpolating
            d_lightstylevalue[j] = 22 * (cl_lightstyle[j].map[0] - 'a');
            continue;
        }

        // interpolate animating light
        // frame just gone
        k = flight % cl_lightstyle[j].length;
        k = cl_lightstyle[j].map[k] - 'a';
        l = (float) (k * 22) * backlerp;

        // upcoming frame
        k = clight % cl_lightstyle[j].length;
        k = cl_lightstyle[j].map[k] - 'a';
        l += (float) (k * 22) * lerpfrac;

        d_lightstylevalue[j] = (int)l;
    }
}

/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

mplane_t	*lightplane;
vec3_t		lightspot;

int RecursiveLightPoint (mnode_t *node, vec3_t start, vec3_t end)
{
    int			r;
    float		front, back, frac;
    int			side;
    mplane_t	*plane;
    vec3_t		mid;
    msurface_t	*surf;
    int			s, t, ds, dt;
    int			i;
    mtexinfo_t	*tex;
    byte		*lightmap;
    unsigned	scale;
    int			maps;

    if (node->contents < 0)
    {
        return -1;		// didn't hit anything
    }

// calculate mid point
    plane = node->plane;
    front = DotProduct (start, plane->normal) - plane->dist;
    back = DotProduct (end, plane->normal) - plane->dist;
    side = front < 0;

    if ( (back < 0) == side)
    {
        return RecursiveLightPoint (node->children[side], start, end);
    }
    frac = front / (front-back);

    mid[0] = start[0] + (end[0] - start[0])*frac;
    mid[1] = start[1] + (end[1] - start[1])*frac;
    mid[2] = start[2] + (end[2] - start[2])*frac;

// go down front side
    r = RecursiveLightPoint (node->children[side], start, mid);

    if (r >= 0)
    {
        return r;		// hit something
    }

    if ( (back < 0) == side )
    {
        return -1;		// didn't hit anuthing
    }

// check for impact on this node
    VectorCopy (mid, lightspot);
    lightplane = plane;

    for (surf = cl.worldmodel->surfaces + node->firstsurface, i=0 ; i<node->numsurfaces ; i++, surf++)
    {
        if (surf->flags & SURF_DRAWTILED) continue;	// no lightmaps

        tex = surf->texinfo;

        s = DotProduct (mid, tex->vecs[0]) + tex->vecs[0][3];
        t = DotProduct (mid, tex->vecs[1]) + tex->vecs[1][3];;

        if (s < surf->texturemins[0] || t < surf->texturemins[1]) continue;

        ds = s - surf->texturemins[0];
        dt = t - surf->texturemins[1];

        if ( ds > surf->extents[0] || dt > surf->extents[1] ) continue;

        if (!surf->samples)
        {
            return 0;
        }
        ds >>= 4;
        dt >>= 4;

        lightmap = surf->samples;
        r = 0;

        if (lightmap)
        {

            lightmap += dt * ((surf->extents[0]>>4)+1) + ds;

            for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ; maps++)
            {
                scale = d_lightstylevalue[surf->styles[maps]];
                r += *lightmap * scale;
                lightmap += ((surf->extents[0]>>4)+1) * ((surf->extents[1]>>4)+1);
            }
            r >>= 8;
        }
        return r;
    }

// go down back side
    return RecursiveLightPoint (node->children[!side], mid, end);
}

int R_LightPoint (vec3_t p)
{
    vec3_t		end;
    int			r;

    if (!cl.worldmodel->lightdata)
    {
        return 255;
    }
    end[0] = p[0];
    end[1] = p[1];
    end[2] = p[2] - 2048;

    r = RecursiveLightPoint (cl.worldmodel->nodes, p, end);

    if (r == -1)
    {
        r = 0;
    }
    return r;
}

